@@ -5,7 +5,7 @@ progress.
 
 import os
 
-from wandb.apis import CommError
+from wandb.errors.error import CommError
 
 
 class Progress(object):
@@ -14,7 +14,10 @@ class Progress(object):
     def __init__(self, file, callback=None):
         self.file = file
         if callback is None:
-            def callback(bites, total): return (bites, total)
+
+            def callback(bites, total):
+                return (bites, total)
+
         self.callback = callback
         self.bytes_read = 0
         self.len = os.fstat(file.fileno()).st_size
@@ -28,7 +31,11 @@ class Progress(object):
             # we could avoid those by updating the self.len in real-time, but
             # files getting truncated while uploading seems like something
             # that shouldn't really be happening anyway.
-            raise CommError('File {} size shrank from {} to {} while it was being uploaded.'.format(self.file.name, self.len, self.bytes_read))
+            raise CommError(
+                "File {} size shrank from {} to {} while it was being uploaded.".format(
+                    self.file.name, self.len, self.bytes_read
+                )
+            )
         # Growing files are also likely to be bad, but our code didn't break
         # on those in the past so it's riskier to make that an error now.
         self.callback(len(bites), self.bytes_read)
